/*
 * Copyright (C) 2016  Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "pathbar.h"
#include "pathbar_p.h"
#include <QToolButton>
#include <QScrollArea>
#include <QScrollBar>
#include <QHBoxLayout>
#include <QResizeEvent>
#include <QContextMenuEvent>
#include <QMenu>
#include <QClipboard>
#include <QApplication>
#include <QDialog>
#include <QLabel>
#include <QTimer>
#include <QPainter>
#include <QDebug>
#include <QStandardPaths>
#include <gio/gfile.h>
#include <QBitmap>
#include "fileoperation.h"
#include "pathedit.h"
#include "utilities.h"
#include "core/bookmarks.h"
#include <QGraphicsDropShadowEffect>
#include <QtDBus/QDBusInterface>

namespace Fm {

class TipDialog : public QDialog {
public:
    explicit TipDialog(QWidget* parent = 0, Qt::WindowFlags f = 0) : QDialog(parent, f) {
        setWindowFlag(Qt::FramelessWindowHint);
        setWindowFlag(Qt::Tool);
        setWindowFlag(Qt::SubWindow);
        setWindowFlag(Qt::Popup);
        setWindowFlag(Qt::X11BypassWindowManagerHint);
        QVBoxLayout *layout = new QVBoxLayout(this);
        QLabel *label = new QLabel(this);
        setStyleSheet("border:1px solid rgb(221, 221, 222);background-color:#ffffff");
        label->setStyleSheet("background-color:transparent;border:none");
        label->setText("收藏成功!");
        layout->addWidget(label);
        setLayout(layout);
        setFixedHeight(40);

        QTimer::singleShot(500, this, &TipDialog::closedialog);
    }
    void closedialog(){
        close();
    }
    void paintEvent(QPaintEvent *event)
    {
        Q_UNUSED(event);
        QBitmap bmp(this->size());
        bmp.fill();
        QPainter p(&bmp);
        p.setPen(QColor(221, 221, 222));
        p.setBrush(Qt::black);
        p.drawRoundedRect(bmp.rect(),5,5);
        setMask(bmp);
    }

};

PathBar::PathBar(QWidget* parent):
    QWidget(parent),
    tempPathEdit_(nullptr),
    toggledBtn_(nullptr),
    Rightbtn_(nullptr) {

    qInfo()<<"PathBar Init";
    QHBoxLayout* topLayout = new QHBoxLayout(this);
    topLayout->setContentsMargins(0, 0, 0, 0);
    topLayout->setSpacing(0);
    bool rtl(layoutDirection() == Qt::RightToLeft);

    // the arrow button used to scroll to start of the path
    scrollToStart_ = new QToolButton(this);
    scrollToStart_->setArrowType(rtl ? Qt::RightArrow : Qt::LeftArrow);
    scrollToStart_->setAutoRepeat(true);
    scrollToStart_->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::MinimumExpanding);
    connect(scrollToStart_, &QToolButton::clicked, this, &PathBar::onScrollButtonClicked);
    topLayout->addWidget(scrollToStart_);

    // there might be too many buttons when the path is long, so make it scrollable.
    scrollArea_ = new QScrollArea(this);
    scrollArea_->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    scrollArea_->setFrameShape(QFrame::NoFrame);
    scrollArea_->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    scrollArea_->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    scrollArea_->setSizeAdjustPolicy(QAbstractScrollArea::AdjustToContents);
    scrollArea_->verticalScrollBar()->setDisabled(true);
    connect(scrollArea_->horizontalScrollBar(), &QAbstractSlider::valueChanged, this, &PathBar::setArrowEnabledState);
    topLayout->addWidget(scrollArea_, 1); // stretch factor=1, make it expandable

    btnbookmark = new QToolButton();
    btnbookmark->setStyleSheet("background-color:transparent;border:none");
//    QPixmap mapimg = QPixmap(":/res/bookmark-click.png");
    btnbookmark->setIcon(QPixmap(":/res/bookmark-click.png"));
    connect(btnbookmark, &QToolButton::clicked, this, &PathBar::onBookmarkClicked);
    topLayout->addWidget(btnbookmark);

    // the arrow button used to scroll to end of the path
    scrollToEnd_ = new QToolButton(this);
    scrollToEnd_->setArrowType(rtl ? Qt::LeftArrow : Qt::RightArrow);
    scrollToEnd_->setAutoRepeat(true);
    scrollToEnd_->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::MinimumExpanding);
    connect(scrollToEnd_, &QToolButton::clicked, this, &PathBar::onScrollButtonClicked);
    topLayout->addWidget(scrollToEnd_);

    // container widget of the path buttons
    buttonsWidget_ = new QWidget(this);
    buttonsWidget_->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    buttonsWidget_->setStyleSheet("QWidget QToolButton {background-color:#ffffff;}");
    //buttonsWidget_->setStyleSheet("QWidget QToolButton {background-color:#ffffff;}");
    setStyleSheet("background-color:transparent;");
    buttonsLayout_ = new QHBoxLayout(buttonsWidget_);
    buttonsLayout_->setContentsMargins(0, 0, 5, 0);
    buttonsLayout_->setSpacing(2);
    buttonsLayout_->setSizeConstraint(QLayout::SetFixedSize); // required when added to scroll area according to QScrollArea doc.
    scrollArea_->setWidget(buttonsWidget_); // make the buttons widget scrollable if the path is too long
}

void PathBar::resizeEvent(QResizeEvent* event) {
    QWidget::resizeEvent(event);
    updateScrollButtonVisibility();
    QTimer::singleShot(0, this, SLOT(ensureToggledVisible()));
}

void PathBar::wheelEvent(QWheelEvent* event) {
    QWidget::wheelEvent(event);
    QAbstractSlider::SliderAction action = QAbstractSlider::SliderNoAction;
    int vDelta = event->angleDelta().y();
    if(vDelta > 0) {
        if(scrollToStart_->isEnabled()) {
            action = QAbstractSlider::SliderSingleStepSub;
        }
    }
    else if(vDelta < 0) {
        if(scrollToEnd_->isEnabled()) {
            action = QAbstractSlider::SliderSingleStepAdd;
        }
    }
    scrollArea_->horizontalScrollBar()->triggerAction(action);
}

void PathBar::mousePressEvent(QMouseEvent* event) {
    QWidget::mousePressEvent(event);
    if(event->button() == Qt::LeftButton) {
        openEditor();
    }
    else if(event->button() == Qt::MiddleButton) {
        PathButton* btn = qobject_cast<PathButton*>(childAt(event->x(), event->y()));
        if(btn != nullptr) {
            scrollArea_->ensureWidgetVisible(btn,
                                             1); // a harmless compensation for a miscalculation in Qt
            Q_EMIT middleClickChdir(pathForButton(btn));
        }
    }
    else if(event->button() == Qt::RightButton) {
        Rightbtn_ = qobject_cast<PathButton*>(childAt(event->x(), event->y()));
    }
}

void PathBar::paintEvent(QPaintEvent *event)
{
    QPainter painter(this);
    painter.setPen(QColor(221, 221, 222));
    painter.setBrush(QColor("#ffffff"));
    painter.drawRect(0,6,this->width()-1,this->height()-13);
//    painter.drawLine(0,5,this->width()-1,5);
//    painter.drawLine(1,5,1,this->height()-6);
//    painter.drawLine(0,this->height()-6,this->width()-1,this->height()-6);
//    painter.drawLine(this->width()-1,5,this->width()-1,this->height()-6);
}

void PathBar::contextMenuEvent(QContextMenuEvent* event) {
    QMenu* menu = new QMenu(this);
    
    connect(menu, &QMenu::aboutToHide, menu, &QMenu::deleteLater);

    QAction* action = menu->addAction(tr("&Edit Path"));
    connect(action, &QAction::triggered, this, &PathBar::openEditor);
    action = menu->addAction(tr("Open in New Tab"));
    connect(action, &QAction::triggered, this, &PathBar::onOpenNewTab);
    action = menu->addAction(tr("Open in New Window"));
    connect(action, &QAction::triggered, this, &PathBar::onOpenNewWindow);
    action = menu->addAction(tr("Cut"));
    action->setIcon(QIcon(":/res/nde-cut.png"));
    connect(action, &QAction::triggered, this, &PathBar::cutPathAndFolder);
    action = menu->addAction(tr("Copy"));
    action->setIcon(QIcon(":/res/nde-copy.png"));
    connect(action, &QAction::triggered, this, &PathBar::copyPathAndFolder);
    action = menu->addAction(tr("Paste to Folder"));
    // 判断是否有粘贴,如果无设置为不可用
    QClipboard* clipboard = QApplication::clipboard();
    const QMimeData* data = clipboard->mimeData();
    Fm::FilePathList paths;
    bool isCut = false;
    std::tie(paths, isCut) = Fm::parseClipboardData(*data);
    if(paths.empty()) {
        action->setEnabled(false);
        action->setIcon(QIcon(":/res/nde-paste-disable.png"));
    }
    else {
        action->setIcon(QIcon(":/res/nde-paste.png"));
    }
    connect(action, &QAction::triggered, this, &PathBar::pasteToFolder);

    // 样式
    menu->setWindowFlag(Qt::FramelessWindowHint);
    menu->setAttribute(Qt::WA_TranslucentBackground);
    bool open = false;
    QDBusInterface * compositer = new QDBusInterface(QString::fromLatin1("org.nde.session"),
                                                     QString::fromLatin1("/NdeSession"),
                                                     QString::fromLatin1("org.nde.session"),
                                                     QDBusConnection::sessionBus());
    if(compositer->isValid())
    {
        QDBusMessage reply = compositer->call("getCompositorOpen");
        if(!reply.arguments().isEmpty()) open = reply.arguments().first().toBool();
    }
    if(open)
    {
        menu->setStyleSheet(QLatin1String("QMenu {background-color: #ffffff; border-radius: 4px; color: #5d626d; margin: 2px;padding: 2px;}"
					  "QMenu::item {padding-left: 17px; padding-right: 35px; height: 32px; font-family: SourceHanSansCN-Normal; font-size: 14px; text-align:left; color: #5d626d; }"
                                          "QMenu::item:!enabled {color: #a8abb1;}"
                                          "QMenu::item:selected{background-color:#2c9dfc; color:#ffffff;}"
                                          "QMenu::icon {margin-left: 10px; margin-right: 10px;}"));
    }
    else
    {
        menu->setStyleSheet(QLatin1String("QMenu {background-color: #ffffff; border: 1px solid #d8d8d8; border-radius: 0px; color: #5d626d; padding: 2px;}"
					  "QMenu::item {padding-left: 17px; padding-right: 35px; height: 32px; font-family: SourceHanSansCN-Normal; font-size: 14px; text-align:left; color: #5d626d; }"
                                          "QMenu::item:!enabled {color: #a8abb1;}"
                                          "QMenu::item:selected{background-color:#2c9dfc; color:#ffffff;}"
                                          "QMenu::icon {margin-left: 10px; margin-right: 10px;}"));
    }

    // 阴影
    QGraphicsDropShadowEffect  * shadow_effect = new QGraphicsDropShadowEffect(this);
    shadow_effect->setOffset(0,0);
    QColor color("#000000");
    color.setAlphaF(0.2);
    shadow_effect->setColor(color);
    shadow_effect->setBlurRadius(4);
    menu->setGraphicsEffect(shadow_effect);

    menu->popup(mapToGlobal(event->pos()));
}

void PathBar::updateScrollButtonVisibility() {
    // Wait for the horizontal scrollbar to be completely shaped.
    // Without this, the enabled state of arrow buttons might be
    // wrong when the pathbar is created for the first time.
    QTimer::singleShot(0, this, SLOT(setScrollButtonVisibility()));
}

void PathBar::setScrollButtonVisibility() {
    bool showScrollers;
    if(tempPathEdit_ != nullptr) {
        showScrollers = false;
    }
    else {
        showScrollers = (buttonsLayout_->sizeHint().width() > width());
    }
    scrollToStart_->setVisible(showScrollers);
    scrollToEnd_->setVisible(showScrollers);
    if(showScrollers) {
        QScrollBar* sb = scrollArea_->horizontalScrollBar();
        int value = sb->value();
        scrollToStart_->setEnabled(value != sb->minimum());
        scrollToEnd_->setEnabled(value != sb->maximum());
    }
}

Fm::FilePath PathBar::pathForButton(PathButton* btn) {
    std::string fullPath;
    int buttonCount = buttonsLayout_->count() - 1; // the last item is a spacer
    for(int i = 0; i < buttonCount; ++i) {
        if(!fullPath.empty() && fullPath.back() != '/') {
            fullPath += '/';
        }
        PathButton* elem = static_cast<PathButton*>(buttonsLayout_->itemAt(i)->widget());
        fullPath += elem->name();
        if(elem == btn)
            break;
    }
    return Fm::FilePath::fromPathStr(fullPath.c_str());
}

void PathBar::setpathbookmark(const Fm::FilePath& path){

    QString pathstr = path.toString().get();

    if (!path.hasParent() || (pathstr.left(1).compare("/") != 0)){
        btnbookmark->setVisible(false);
        return ;
    } else {
        btnbookmark->setVisible(true);
    }

    if(Bookmarks::globalInstance()->exist(path)){
        inpath = true;
    } else {
        inpath = false;
    }

    if(inpath){
        btnbookmark->setIcon(QPixmap(":/res/bookmark-click.png"));
    } else {
        btnbookmark->setIcon(QPixmap(":/res/bookmark.png"));
    }
}

void PathBar::onButtonToggled(bool checked) {
    if(checked) {
        PathButton* btn = static_cast<PathButton*>(sender());
        toggledBtn_ = btn;
        currentPath_ = pathForButton(btn);
        setpathbookmark(currentPath_);
        Q_EMIT chdir(currentPath_);

        // since scrolling to the toggled buton will happen correctly only when the
        // layout is updated and because the update is disabled on creating buttons
        // in setPath(), the update status can be used as a sign to know when to wait
        if(updatesEnabled()) {
            scrollArea_->ensureWidgetVisible(btn, 1);
        }
        else {
            QTimer::singleShot(0, this, SLOT(ensureToggledVisible()));
        }
    }
}

void PathBar::ensureToggledVisible() {
    if(toggledBtn_ != nullptr && tempPathEdit_ == nullptr) {
        scrollArea_->ensureWidgetVisible(toggledBtn_, 1);
    }
}

void PathBar::onScrollButtonClicked() {
    QToolButton* btn = static_cast<QToolButton*>(sender());
    QAbstractSlider::SliderAction action = QAbstractSlider::SliderNoAction;
    if(btn == scrollToEnd_) {
        action = QAbstractSlider::SliderSingleStepAdd;
    }
    else if(btn == scrollToStart_) {
        action = QAbstractSlider::SliderSingleStepSub;
    }
    scrollArea_->horizontalScrollBar()->triggerAction(action);
}

void PathBar::onBookmarkClicked(){

    if(!currentPath_){
        return ;
    }
    if(inpath){
       Bookmarks::globalInstance()->remove(currentPath_);
       Bookmarks::globalInstance()->DetePathTimer();
    } else {
        Bookmarks::globalInstance()->insert(currentPath_, currentPath_.baseName().get(), -1);

        TipDialog dialog(nullptr);
        QCursor cur;
        dialog.move(cur.pos().x() - 60,cur.pos().y() +10);
        dialog.exec();
    }
    setpathbookmark(currentPath_);
}

void PathBar::setPath(Fm::FilePath path) {
    if(currentPath_ == path) { // same path, do nothing
        return;
    }

    setpathbookmark(path);

    auto oldPath = std::move(currentPath_);
    currentPath_ = std::move(path);
    // check if we already have a button for this path
    int buttonCount = buttonsLayout_->count() - 1; // the last item is a spacer
    if(oldPath && currentPath_.isPrefixOf(oldPath)) {
        for(int i = buttonCount - 1; i >= 0; --i) {
            auto btn = static_cast<PathButton*>(buttonsLayout_->itemAt(i)->widget());
            if(pathForButton(btn) == currentPath_) {
                btn->setChecked(true); // toggle the button
                /* we don't need to emit chdir signal here since later
                 * toggled signal will be triggered on the button, which
                 * in turns emit chdir. */
                return;
            }
        }
    }

    /* FIXME: if the new path is the subdir of our full path, actually
     *        we can append several new buttons rather than re-create
     *        all of the buttons. This can reduce flickers. */

    setUpdatesEnabled(false);
    toggledBtn_ = nullptr;
    // we do not have the path in the buttons list
    // destroy existing path element buttons and the spacer
    QLayoutItem* item;
    while((item = buttonsLayout_->takeAt(0)) != nullptr) {
        delete item->widget();
        delete item;
    }

    // create new buttons for the new path
    auto btnPath = currentPath_;


    auto filepath = Bookmarks::globalInstance()->getBookmarkPath();

    bool showother = true;
    while(btnPath) {
        Fm::CStrPtr name;
        Fm::CStrPtr displayName;
        auto parent = btnPath.parent();
        // FIXME: some buggy uri types, such as menu://, fail to return NULL when there is no parent path.
        // Instead, the path itself is returned. So we check if the parent path is the same as current path.
        auto isRoot = !parent.isValid() || parent == btnPath;
        if(isRoot) {
            displayName = btnPath.displayName();
            name = btnPath.toString();
        }
        else {
            name = btnPath.baseName();
        }
        PathButton * btn;
        if(btnPath == filepath){
            btn = new PathButton(name.get(), tr("BookMark"), isRoot, buttonsWidget_);
        } else {
            btn = new PathButton(name.get(), displayName ? displayName.get() : name.get(), isRoot, buttonsWidget_);
        }

        btn->setFixedHeight(28);
        btn->show();
        if(!showother){
            btn->setVisible(false);
        }
        connect(btn, &QAbstractButton::toggled, this, &PathBar::onButtonToggled);
        buttonsLayout_->insertWidget(0, btn);
        if(isRoot) { // this is the root element of the path
            break;
        }
        if(btnPath == filepath){
            showother = false;
        }
        btnPath = parent;
    }

    buttonsLayout_->addStretch(0); // add a spacer at the tail of the buttons



    // we don't want to scroll vertically. make the scroll area fit the height of the buttons
    // FIXME: this is a little bit hackish :-(
    scrollArea_->setFixedHeight(buttonsLayout_->sizeHint().height());
    updateScrollButtonVisibility();

    // to guarantee that the button will be scrolled to correctly,
    // it should be toggled only after the layout update starts above
    buttonCount = buttonsLayout_->count() - 1;
    if(buttonCount > 0) {
        PathButton* lastBtn = static_cast<PathButton*>(buttonsLayout_->itemAt(buttonCount - 1)->widget());
        // we don't have to emit the chdir signal since the "onButtonToggled()" slot will be triggered by this.
        lastBtn->setChecked(true);
    }

    setUpdatesEnabled(true);
}

void PathBar::openEditor() {
    if(tempPathEdit_ == nullptr) {
        tempPathEdit_ = new PathEdit(this);
        delete layout()->replaceWidget(scrollArea_, tempPathEdit_, Qt::FindDirectChildrenOnly);
        scrollArea_->hide();
        scrollToStart_->setVisible(false);
        scrollToEnd_->setVisible(false);
        btnbookmark->setVisible(false);
        tempPathEdit_->setText(currentPath_.toString().get());

        connect(tempPathEdit_, &PathEdit::returnPressed, this, &PathBar::onReturnPressed);
        connect(tempPathEdit_, &PathEdit::editingFinished, this, &PathBar::closeEditor);
    }
    tempPathEdit_->selectAll();
    QApplication::clipboard()->setText(tempPathEdit_->text(), QClipboard::Selection);
    QTimer::singleShot(0, tempPathEdit_, SLOT(setFocus()));
}

void PathBar::closeEditor() {
    if(tempPathEdit_ == nullptr) {
        return;
    }
    // If a menu has popped up synchronously (with QMenu::exec), the path buttons may be drawn
    // but the path-edit may not disappear until the menu is closed. So, we hide it here.
    // However, since hiding the path-edit makes it lose focus and emit editingFinished(),
    // we should first disconnect from it to avoid recursive calling of the current function.
    tempPathEdit_->disconnect();
    tempPathEdit_->setVisible(false);
    delete layout()->replaceWidget(tempPathEdit_, scrollArea_, Qt::FindDirectChildrenOnly);
    scrollArea_->show();
    if(buttonsLayout_->sizeHint().width() > width()) {
        scrollToStart_->setVisible(true);
        scrollToEnd_->setVisible(true);
    }

    tempPathEdit_->deleteLater();
    tempPathEdit_ = nullptr;
    btnbookmark->setVisible(true);
    updateScrollButtonVisibility();

    Q_EMIT editingFinished();
}

void PathBar::copyPath() {
    QApplication::clipboard()->setText(currentPath_.toString().get());
}

void PathBar::onReturnPressed() {
    QByteArray pathStr = tempPathEdit_->text().toLocal8Bit();
    setPath(Fm::FilePath::fromPathStr(pathStr.constData()));
}

void PathBar::setArrowEnabledState(int value) {
    if(buttonsLayout_->sizeHint().width() > width()) {
        QScrollBar* sb = scrollArea_->horizontalScrollBar();
        scrollToStart_->setEnabled(value != sb->minimum());
        scrollToEnd_->setEnabled(value != sb->maximum());
    }
}

void PathBar::onOpenNewTab() {
    if(Rightbtn_ != nullptr) {
        Q_EMIT chdirRequested(1, pathForButton(Rightbtn_));
    }
}

void PathBar::onOpenNewWindow() {
    if(Rightbtn_ != nullptr) {
        Q_EMIT chdirRequested(2, pathForButton(Rightbtn_));
    }
}

void PathBar::copyPathAndFolder() {
    Fm::FilePath selectPath = pathForButton(Rightbtn_);
    // 只复制路径
    QApplication::clipboard()->setText(selectPath.toString().get());
    // 复制文件夹
    Fm::FilePathList paths;
    paths.push_back(selectPath);
    Fm::copyFilesToClipboard(paths);
}

void PathBar::cutPathAndFolder() {
    Fm::FilePath selectPath = pathForButton(Rightbtn_);
    // 只剪切路径
    QApplication::clipboard()->setText(selectPath.toString().get());
    // 剪切文件夹
    Fm::FilePathList paths;
    paths.push_back(selectPath);
    Fm::cutFilesToClipboard(paths);
}

void PathBar::pasteToFolder() {
    Fm::FilePath selectPath = pathForButton(Rightbtn_);
    // 粘贴到文件夹
    pasteFilesFromClipboard(selectPath);
}

} // namespace Fm
