/*
 * Copyright (C) 2012 - 2015  Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


#include "placesview.h"
#include "placesmodel.h"
#include "placesmodelitem.h"
#include "mountoperation.h"
#include "fileoperation.h"
#include <QMenu>
#include <QContextMenuEvent>
#include <QHeaderView>
#include <QDebug>
#include <QGuiApplication>
#include "folderitemdelegate.h"
#include <QGraphicsDropShadowEffect>
#include <QtDBus/QDBusInterface>

namespace Fm {



PlacesProxyModel::PlacesProxyModel(QObject* parent) :
    QSortFilterProxyModel(parent),
    showAll_(false),
    hiddenItemsRestored_(false) {
}

PlacesProxyModel::~PlacesProxyModel() {
}

void PlacesProxyModel::restoreHiddenItems(const QSet<QString>& items) {
    // hidden items should be restored only once
    if(!hiddenItemsRestored_ && !items.isEmpty()) {
        hidden_.clear();
        QSet<QString>::const_iterator i = items.constBegin();
        while (i != items.constEnd()) {
            if(!(*i).isEmpty()) {
                hidden_ << *i;
            }
            ++i;
        }
        hiddenItemsRestored_ = true;
        invalidateFilter();
    }
}

void PlacesProxyModel::setHidden(const QString& str, bool hide) {
    if(hide) {
        if(!str.isEmpty()) {
            hidden_ << str;
        }
    }
    else {
        hidden_.remove(str);
    }
    invalidateFilter();
}

void PlacesProxyModel::showAll(bool show) {
    showAll_ = show;
    invalidateFilter();
}

bool PlacesProxyModel::filterAcceptsRow(int source_row, const QModelIndex& source_parent) const {
    if(showAll_ || hidden_.isEmpty()) {
        return true;
    }
    if(PlacesModel* srcModel = static_cast<PlacesModel*>(sourceModel())) {
        QModelIndex index = srcModel->index(source_row, 0, source_parent);
        if(PlacesModelItem* item = static_cast<PlacesModelItem*>(srcModel->itemFromIndex(index))) {
            if(item->type() == PlacesModelItem::Places) {
                if(auto path = item->path()) {
                    if(hidden_.contains(path.toString().get())) {
                        return false;
                    }
                }
            }
            else if(item->type() == PlacesModelItem::Volume) {
                CStrPtr uuid{g_volume_get_uuid(static_cast<PlacesModelVolumeItem*>(item)->volume())};
                if(uuid && hidden_.contains(uuid.get())) {
                    return false;
                }
            }
            // show a root items only if, at least, one of its children is shown
            else if((source_row == 0 || source_row == 1) && !source_parent.isValid()) {
                QModelIndex indx = index.child(0, 0);
                while(PlacesModelItem* childItem = static_cast<PlacesModelItem*>(srcModel->itemFromIndex(indx))) {
                    if(childItem->type() == PlacesModelItem::Places) {
                        if(auto path = childItem->path()) {
                            if(!hidden_.contains(path.toString().get())) {
                                return true;
                            }
                        }
                    }
                    else if(childItem->type() == PlacesModelItem::Volume) {
                        CStrPtr uuid{g_volume_get_uuid(static_cast<PlacesModelVolumeItem*>(childItem)->volume())};
                        if(uuid == nullptr || !hidden_.contains(uuid.get())) {
                            return true;
                        }
                    }
                    else {
                        return true;
                    }
                    indx = indx.sibling(indx.row() + 1, 0);
                }
                return false;
            }
        }
    }
    return true;
}

PlacesView::PlacesView(QWidget* parent):
    QTreeView(parent) {
    setRootIsDecorated(true);
    setAlternatingRowColors(true);
    setHeaderHidden(true);
    setIndentation(15);
    setStyleSheet("QTreeView {background-color:#f6f9fb; border:1px solid #dae2e9;}"
                  "QTreeView::item{height:30px;color:#6c727a;}"
                  "QTreeView::item:selected {background-color:#2c9dfc;color:#ffffff;}");
    connect(this, &QTreeView::clicked, this, &PlacesView::onClicked);
    connect(this, &QTreeView::pressed, this, &PlacesView::onPressed);

    setIconSize(QSize(24, 24));

    FolderItemDelegate* delegate = new FolderItemDelegate(this, this);
    delegate->setFileInfoRole(PlacesModel::FileInfoRole);
    delegate->setIconInfoRole(PlacesModel::FmIconRole);
    setItemDelegateForColumn(0, delegate);

    model_ = std::make_shared<PlacesModel>();
    //if(!proxyModel_) {
        proxyModel_ = std::make_shared<PlacesProxyModel>();
    //}
    //if(!proxyModel_->sourceModel()) { // all places-views may have been closed
        proxyModel_->setSourceModel(model_.get());
    //}
    setModel(proxyModel_.get());

    // these 2 connections are needed to update filtering
    connect(model_.get(), &QAbstractItemModel::rowsInserted, this, [this](const QModelIndex&, int, int) {
        proxyModel_->setHidden(QString()); // just invalidates filter
        expandAll();
        // for some reason (a Qt bug?), spanning is reset
        spanFirstColumn();
    });
    connect(model_.get(), &QAbstractItemModel::rowsRemoved, this, [this](const QModelIndex&, int, int) {
        proxyModel_->setHidden(QString());
        updateFocus();
    });

    QHeaderView* headerView = header();
    headerView->setSectionResizeMode(0, QHeaderView::Stretch);
    headerView->setSectionResizeMode(1, QHeaderView::Fixed);
    headerView->setStretchLastSection(false);
    expandAll();

    spanFirstColumn();

    // the 2nd column is for the eject buttons
    setSelectionBehavior(QAbstractItemView::SelectRows); // FIXME: why this does not work?
    setAllColumnsShowFocus(false);

    setAcceptDrops(false);
    setDragEnabled(false);

    // update the umount button's column width based on icon size
    onIconSizeChanged(iconSize());
#if QT_VERSION >= QT_VERSION_CHECK(5, 5, 0) // this signal requires Qt >= 5.5
    connect(this, &QAbstractItemView::iconSizeChanged, this, &PlacesView::onIconSizeChanged);
#endif
}

PlacesView::~PlacesView() {
    // qDebug("delete PlacesView");
}

void PlacesView::spanFirstColumn() {
    // FIXME: is there any better way to make the first column span the whole row?
    setFirstColumnSpanned(0, QModelIndex(), true); // places root
    setFirstColumnSpanned(1, QModelIndex(), true); // devices root
    setFirstColumnSpanned(2, QModelIndex(), true); // network root
    // NOTE: The first column of the devices children shouldn't be spanned
    // because the second column contains eject buttons.
    QModelIndex indx = proxyModel_->mapFromSource(model_->placesRoot->index());
    if(indx.isValid()) {
        for(int i = 0; i < indx.model()->rowCount(indx); ++i) {
            setFirstColumnSpanned(i, indx, true);
        }
    }
    indx = proxyModel_->mapFromSource(model_->networkRoot->index());
    if(indx.isValid()) {
        for(int i = 0; i < indx.model()->rowCount(indx); ++i) {
            setFirstColumnSpanned(i, indx, true);
        }
    }

    indx = proxyModel_->mapFromSource(model_->devicesRoot->index());
    if(indx.isValid()) {
        qInfo()<<"rowCount:"<<indx.model()->rowCount(indx);
        for(int i = 0; i < indx.model()->rowCount(indx); ++i) {
            QModelIndex itemIndex = indx.child(i,0);
            PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(itemIndex)));
            if(item->type() != PlacesModelItem::Volume)
            {
                setFirstColumnSpanned(i, indx, true);
            }
            else
            {
                setFirstColumnSpanned(i, indx, false);
            }
        }
    }
}

void PlacesView::activateRow(int type, const QModelIndex& index) {
    /*if(!index.parent().isValid()) { // ignore root items
        return;
    }*/
    PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(index)));
    if(item) {

        if (preItem && (preItem != item)){
            updatePreItemIcon();
        }

        item->updateClickIcon();
        if (item->type() == PlacesModelItem::Volume)
        {
            PlacesModelVolumeItem* volumeItem = static_cast<PlacesModelVolumeItem*>(item);
            if(volumeItem->isMounted() && volumeItem->volumeFromUSB())
            {
                QModelIndex btnIndex = index.sibling(index.row(), 1); // the real eject button is at column 1
                QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                ejectBtn->setIcon(QIcon(":/res/eject-click.png"));
            }
        }
        preItem = item;
        auto path = item->path();
        if(!path) {
            // check if mounting volumes is needed
            if(item->type() == PlacesModelItem::Volume) {
                PlacesModelVolumeItem* volumeItem = static_cast<PlacesModelVolumeItem*>(item);
                if(!volumeItem->isMounted() && !volumeItem->getEject()) {
                    // Mount the volume
                    GVolume* volume = volumeItem->volume();
                    MountOperation* op = new MountOperation(true, this);
                    op->mount(volume);
                    // connect(op, SIGNAL(finished(GError*)), SLOT(onMountOperationFinished(GError*)));
                    // blocking here until the mount operation is finished?

                    // FIXME: update status of the volume after mount is finished!!
                    if(!op->wait()) {
                        return;
                    }
                    path = item->path();
                }
            }
        }
        if(path && path.isExists()) {
            Q_EMIT chdirRequested(type, path);
        } else {
            if(item->type() == PlacesModelItem::Type::NetServer){
                Q_EMIT connetServer();
            }
        }
    }
}

void PlacesView::activateAfterEject(int type, const QModelIndex &index)
{
    PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(index)));
    if(item)
    {
        if(preItem && (preItem != item))
        {
            auto path = preItem->path();
            if(path)
            {
                QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                setFocus();
                setCurrentIndex(preIndex);
                Q_EMIT chdirRequested(type, path);
            }
        }
        else
        {
            QModelIndex nextIndex = index.parent();
            PlacesModelItem* nextItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(nextIndex)));
            auto path = nextItem->path();
            if(path)
            {
                setFocus();
                setCurrentIndex(nextIndex);
                Q_EMIT chdirRequested(type, path);
            }
        }
    }
}

// mouse button pressed
void PlacesView::onPressed(const QModelIndex& index) {
    // if middle button is pressed
    if(QGuiApplication::mouseButtons() & Qt::MiddleButton) {
        // the real item is at column 0
        activateRow(1, 0 == index.column() ? index : index.sibling(index.row(), 0));
    }
}

void PlacesView::onIconSizeChanged(const QSize& size) {
    setColumnWidth(1, size.width() + 5);
}

void PlacesView::onEjectButtonClicked(PlacesModelItem* item) {
    // The eject button is clicked for a device item (volume or mount)
    if(item->type() == PlacesModelItem::Volume)
    {
        QModelIndex parentIndx = proxyModel_->mapFromSource(model_->devicesRoot->index());

        PlacesModelVolumeItem* volumeItem = static_cast<PlacesModelVolumeItem*>(item);
        if(volumeItem->canEject()) { // do eject if applicable
            volumeItem->setEject(true);
            MountOperation* op = new MountOperation(true, this);
            op->eject(volumeItem->volume());
            op->wait();
        }
        else
        { // otherwise, do unmount instead
            // drive with more than one volume
            PlacesModelItem* tempItem = nullptr;
            GList *ll,*volumes;
            GDrive *drive = g_volume_get_drive(volumeItem->volume());
            gchar *device_name = g_volume_get_identifier(volumeItem->volume(), G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
            char *command = g_strdup_printf("udisksctl power-off -b %s", device_name);
            if (drive)
            {
                volumes = g_drive_get_volumes(drive);       //get本驱动器上的所有卷设备
                if(nullptr != volumes)
                {
                    for(ll = volumes; ll; ll = ll->next){ //遍历驱动器上的所有卷设备
                        GVolume* volume = G_VOLUME(ll->data);

                        /**********/
                        for(int i = 0; i < parentIndx.model()->rowCount(parentIndx); ++i) {
                            QModelIndex tempItemIndex = parentIndx.child(i,0);
                            tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(tempItemIndex)));
                            if(tempItem->type() == PlacesModelItem::Volume)
                            {
                                PlacesModelVolumeItem* tempVolumeItem = static_cast<PlacesModelVolumeItem*>(tempItem);
                                if(tempVolumeItem->volume() == volume)
                                {
                                    tempVolumeItem->setEject(true);
                                    break;
                                }
                            }
                        }
                        /**********/

                        char *umountCommand = g_strdup_printf("umount %s", g_volume_get_identifier(volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE));
                        system(umountCommand);
                        if(tempItem == preItem)
                            preItem = nullptr;
                        g_object_unref(volume);
                    }//for
                    g_list_free(volumes);
                }
                g_object_unref(drive);    //GList对象释放函数
            }
            system(command);
        }
    }
    if(preItem == item){
        preItem = nullptr;
    }
    qDebug("PlacesView::onEjectButtonClicked");
}

void PlacesView::onClicked(const QModelIndex& index) {
    /*if(!index.parent().isValid()) { // ignore root items
        return;
    }*/

    if(index.column() == 0) {
        activateRow(0, index);
    }
    else if(index.column() == 1) { // column 1 contains eject buttons of the mounted devices
        if(index.parent() == proxyModel_->mapFromSource(model_->devicesRoot->index())) { // this is a mounted device
            // the eject button is clicked
            QModelIndex itemIndex = index.sibling(index.row(), 0); // the real item is at column 0
            PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(itemIndex)));
            if(item->type() == PlacesModelItem::Volume) {
                // eject the volume or the mount
                onEjectButtonClicked(item);
                activateAfterEject(0, itemIndex);
            }
            else
            {
                activateRow(0, itemIndex);
            }
        }
        else {
            activateRow(0, index.sibling(index.row(), 0));
        }
    }
}

void PlacesView::setCurrentPath(Fm::FilePath path) {
    currentPath_ = std::move(path);
    if(currentPath_) {
        // TODO: search for item with the path in model_ and select it.
        PlacesModelItem* item = model_->itemFromPath(currentPath_);
        if(item) {
            selectionModel()->setCurrentIndex(proxyModel_->mapFromSource(item->index()), QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);
            if (preItem && (preItem != item)){
                updatePreItemIcon();
            }
            item->updateClickIcon();
            if(item->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* itemVolume = static_cast<PlacesModelVolumeItem*>(item);
                if(itemVolume->isMounted() && itemVolume->volumeFromUSB())
                {
                    QModelIndex itemIndex = proxyModel_->mapFromSource(item->index());
                    QModelIndex btnIndex = itemIndex.sibling(itemIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-click.png"));
                }
            }
            preItem = item;
        }
        else {
            clearSelection();
            if (preItem ){
                preItem->updateIcon();
                if (preItem->type() == PlacesModelItem::Volume)
                {
                    PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                    if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                    {
                        QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                        QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                        QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                        ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                    }
                }
                preItem = item;
            }
        }
    }
    else {
        clearSelection();
    }
}

void PlacesView::updateFocus()
{
    if(currentPath_)
    {
        PlacesModelItem* item = model_->itemFromPath(currentPath_);
        if(item)
        {
            selectionModel()->setCurrentIndex(proxyModel_->mapFromSource(item->index()), QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);


            item->updateClickIcon();
            if(item->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* itemVolume = static_cast<PlacesModelVolumeItem*>(item);
                if(itemVolume->isMounted() && itemVolume->volumeFromUSB())
                {
                    QModelIndex itemIndex = proxyModel_->mapFromSource(item->index());
                    QModelIndex btnIndex = itemIndex.sibling(itemIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-click.png"));
                }
            }
            preItem = item;
        }
    }
    else
    {
        preItem = nullptr;
        clearSelection();
    }
}

void PlacesView::updatePreItemIcon()
{
    QModelIndex indx = proxyModel_->mapFromSource(model_->placesRoot->index());
    if(indx.isValid()) {
        PlacesModelItem* tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(indx)));
        if (tempItem == preItem){
            preItem->updateIcon();
            if (preItem->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                {
                    QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                    QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                }
            }
        }

        for(int i = 0; i < indx.model()->rowCount(indx); ++i) {
            QModelIndex itemIndex = indx.child(i,0);
            PlacesModelItem* tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(itemIndex)));
            if (tempItem == preItem){
                preItem->updateIcon();
                if (preItem->type() == PlacesModelItem::Volume)
                {
                    PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                    if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                    {
                        QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                        QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                        QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                        ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                    }
                }
            }
        }
    }

    indx = proxyModel_->mapFromSource(model_->devicesRoot->index());
    if(indx.isValid()) {
        PlacesModelItem* tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(indx)));
        if (tempItem == preItem){
            preItem->updateIcon();
            if (preItem->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                {
                    QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                    QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                }
            }
        }

        for(int i = 0; i < indx.model()->rowCount(indx); ++i) {
            QModelIndex itemIndex = indx.child(i,0);
            PlacesModelItem* tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(itemIndex)));
            if (tempItem == preItem){
                preItem->updateIcon();
                if (preItem->type() == PlacesModelItem::Volume)
                {
                    PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                    if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                    {
                        QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                        QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                        QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                        ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                    }
                }
            }
        }
    }

    indx = proxyModel_->mapFromSource(model_->networkRoot->index());
    if(indx.isValid()) {
        PlacesModelItem* tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(indx)));
        if (tempItem == preItem){
            preItem->updateIcon();
            if (preItem->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                {
                    QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                    QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                }
            }
        }

        for(int i = 0; i < indx.model()->rowCount(indx); ++i) {
            QModelIndex itemIndex = indx.child(i,0);
            PlacesModelItem* tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(itemIndex)));
            if (tempItem == preItem){
                preItem->updateIcon();
                if (preItem->type() == PlacesModelItem::Volume)
                {
                    PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                    if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                    {
                        QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                        QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                        QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                        ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                    }
                }
            }
        }
    }
}

void PlacesView::dragMoveEvent(QDragMoveEvent* event) {
    QTreeView::dragMoveEvent(event);
    /*
    QModelIndex index = indexAt(event->pos());
    if(event->isAccepted() && index.isValid() && index.parent() == model_->bookmarksRoot->index()) {
      if(dropIndicatorPosition() != OnItem) {
        event->setDropAction(Qt::LinkAction);
        event->accept();
      }
    }
    */
}

void PlacesView::dropEvent(QDropEvent* event) {
    QTreeView::dropEvent(event);
}

void PlacesView::onEmptyTrash() {
    Fm::FilePathList files;
    files.push_back(Fm::FilePath::fromUri("trash:///"));
    Fm::FileOperation::deleteFiles(std::move(files), true, this);
}

void PlacesView::onMoveBookmarkUp() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelBookmarkItem* item = static_cast<PlacesModelBookmarkItem*>(model_->itemFromIndex(action->index()));

    int row = item->row();
    if(row > 0) {
        auto bookmarkItem = item->bookmark();
        Fm::Bookmarks::globalInstance()->reorder(bookmarkItem, row - 1);
    }
}

void PlacesView::onMoveBookmarkDown() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelBookmarkItem* item = static_cast<PlacesModelBookmarkItem*>(model_->itemFromIndex(action->index()));

    int row = item->row();
    if(row < model_->rowCount()) {
        auto bookmarkItem = item->bookmark();
        Fm::Bookmarks::globalInstance()->reorder(bookmarkItem, row + 1);
    }
}

void PlacesView::onDeleteBookmark() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelBookmarkItem* item = static_cast<PlacesModelBookmarkItem*>(model_->itemFromIndex(action->index()));
    auto bookmarkItem = item->bookmark();
    Fm::Bookmarks::globalInstance()->remove(bookmarkItem);
}

// virtual
void PlacesView::commitData(QWidget* editor) {
    QTreeView::commitData(editor);
    PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(currentIndex())));
    if(item->type() == PlacesModelItem::Bookmark)
    {
        PlacesModelBookmarkItem* bookmarkItems = static_cast<PlacesModelBookmarkItem*>(item);
        auto bookmarkItem = bookmarkItems->bookmark();
        // rename bookmark
        Fm::Bookmarks::globalInstance()->rename(bookmarkItem, bookmarkItems->text());
    }
}

void PlacesView::onOpenNewTab() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(action->index()));
    if(item) {
        Q_EMIT chdirRequested(1, item->path());
    }
}

void PlacesView::onOpenNewWindow() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(action->index()));
    if(item) {
        Q_EMIT chdirRequested(2, item->path());
    }
}

void PlacesView::onRenameBookmark() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelBookmarkItem* item = static_cast<PlacesModelBookmarkItem*>(model_->itemFromIndex(action->index()));
    setFocus();
    setCurrentIndex(proxyModel_->mapFromSource(item->index()));
    edit(proxyModel_->mapFromSource(item->index()));
}

void PlacesView::onMountVolume() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelVolumeItem* item = static_cast<PlacesModelVolumeItem*>(model_->itemFromIndex(action->index()));
    MountOperation* op = new MountOperation(true, this);
    op->mount(item->volume());
    op->wait();
}

void PlacesView::onUnmountVolume() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelVolumeItem* item = static_cast<PlacesModelVolumeItem*>(model_->itemFromIndex(action->index()));

    QModelIndex nextIndex = proxyModel_->mapFromSource(item->index()).parent();
    PlacesModelItem* nextItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(nextIndex)));
    auto path = nextItem->path();
    Q_EMIT chdirRequested(0, path);

    MountOperation* op = new MountOperation(true, this);
    op->unmount(item->volume());
    op->wait();
}

void PlacesView::onUnmountMount() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelMountItem* item = static_cast<PlacesModelMountItem*>(model_->itemFromIndex(action->index()));
    if(preItem == item){
        preItem = nullptr;
    }
    QModelIndex nextIndex = proxyModel_->mapFromSource(item->index()).parent();
    PlacesModelItem* nextItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(nextIndex)));
    auto path = nextItem->path();
    Q_EMIT chdirRequested(0, path);

    GMount* mount = item->mount();
    MountOperation* op = new MountOperation(true, this);
    op->unmount(mount);
    op->wait();

    Q_EMIT chdirRequested(0, path);
}

void PlacesView::onEjectVolume() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelVolumeItem* item = static_cast<PlacesModelVolumeItem*>(model_->itemFromIndex(action->index()));
    if(preItem == item){
        preItem = nullptr;
    }
    QModelIndex nextIndex = proxyModel_->mapFromSource(item->index()).parent();
    PlacesModelItem* nextItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(nextIndex)));
    auto path = nextItem->path();
    Q_EMIT chdirRequested(0, path);

    MountOperation* op = new MountOperation(true, this);
    QModelIndex parentIndx = proxyModel_->mapFromSource(model_->devicesRoot->index());
    if(item->canEject()) { // do eject if applicable
        item->setEject(true);
        MountOperation* op = new MountOperation(true, this);
        op->eject(item->volume());
        op->wait();
    }
    else
    { // otherwise, do unmount instead
        // drive with more than one volume
        PlacesModelItem* tempItem = nullptr;
        GList *ll,*volumes;
        GDrive *drive = g_volume_get_drive(item->volume());
        gchar *device_name = g_volume_get_identifier(item->volume(), G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
        char *command = g_strdup_printf("udisksctl power-off -b %s", device_name);
        if (drive)
        {
            volumes = g_drive_get_volumes(drive);       //get本驱动器上的所有卷设备
            if(nullptr != volumes)
            {
                for(ll = volumes; ll; ll = ll->next){ //遍历驱动器上的所有卷设备
                    GVolume* volume = G_VOLUME(ll->data);

                    /**********/
                    for(int i = 0; i < parentIndx.model()->rowCount(parentIndx); ++i) {
                        QModelIndex tempItemIndex = parentIndx.child(i,0);
                        tempItem = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(tempItemIndex)));
                        if(tempItem->type() == PlacesModelItem::Volume)
                        {
                            PlacesModelVolumeItem* tempVolumeItem = static_cast<PlacesModelVolumeItem*>(tempItem);
                            if(tempVolumeItem->volume() == volume)
                            {
                                tempVolumeItem->setEject(true);
                                break;
                            }
                        }
                    }
                    /**********/

                    char *umountCommand = g_strdup_printf("umount %s", g_volume_get_identifier(volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE));
                    system(umountCommand);
                    if(tempItem == preItem)
                        preItem = nullptr;
                    g_object_unref(volume);
                }//for
                g_list_free(volumes);
            }
            g_object_unref(drive);    //GList对象释放函数
        }
        system(command);
    }
}

void PlacesView::contextMenuEvent(QContextMenuEvent* event) {
    QModelIndex index = indexAt(event->pos());
    if(index.isValid()) {
        if(index.column() != 0) { // the real item is at column 0
            index = index.sibling(index.row(), 0);
        }

        // Do not take the ownership of the menu since
        // it will be deleted with deleteLater() upon hidden.
        // This is possibly related to #145 - https://github.com/lxqt/pcmanfm-qt/issues/145
        QMenu* menu = new QMenu();
        QAction* action = nullptr;
        PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(proxyModel_->mapToSource(index)));

        if (preItem && (preItem != item)){
            preItem->updateIcon();
            if (preItem->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* preVolumeItem = static_cast<PlacesModelVolumeItem*>(preItem);
                if(preVolumeItem->isMounted() && preVolumeItem->volumeFromUSB())
                {
                    QModelIndex preIndex = proxyModel_->mapFromSource(preItem->index());
                    QModelIndex btnIndex = preIndex.sibling(preIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-normal.png"));
                }
            }
        }
        if(item) {
            item->updateClickIcon();
            if(item->type() == PlacesModelItem::Volume)
            {
                PlacesModelVolumeItem* itemVolume = static_cast<PlacesModelVolumeItem*>(item);
                if(itemVolume->isMounted() && itemVolume->volumeFromUSB())
                {
                    QModelIndex itemIndex = proxyModel_->mapFromSource(item->index());
                    QModelIndex btnIndex = itemIndex.sibling(itemIndex.row(), 1);
                    QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                    ejectBtn->setIcon(QIcon(":/res/eject-click.png"));
                }
            }
            preItem = item;
        }

        // 样式
        menu->setWindowFlag(Qt::FramelessWindowHint);
        menu->setAttribute(Qt::WA_TranslucentBackground);
        bool open = false;
        QDBusInterface * compositer = new QDBusInterface(QString::fromLatin1("org.nde.session"),
                                                         QString::fromLatin1("/NdeSession"),
                                                         QString::fromLatin1("org.nde.session"),
                                                         QDBusConnection::sessionBus());
        if(compositer->isValid())
        {
            QDBusMessage reply = compositer->call("getCompositorOpen");
            if(!reply.arguments().isEmpty()) open = reply.arguments().first().toBool();
        }
        if(open)
        {
            menu->setStyleSheet(QLatin1String("QMenu { border-radius: 4px; color: #5d626d; margin: 2px;padding: 2px;}"
                                              "QMenu::item { padding-left: 34px; padding-right: 46px; height: 32px; font-family: SourceHanSansCN-Normal; font-size: 14px; text-align:left; color: #5d626d; }"
                                              "QMenu::item:!enabled {color: #a8abb1;}"
                                              "QMenu::item:selected {background-color:#2c9dfc; color:#ffffff;}"));
        }
        else
        {
            menu->setStyleSheet(QLatin1String("QMenu {border: 1px solid #d8d8d8; border-radius: 0px; color: #5d626d; padding: 2px;}"
                                              "QMenu::item { padding-left: 34px; padding-right: 46px; height: 32px; font-family: SourceHanSansCN-Normal; font-size: 14px; text-align:left; color: #5d626d; }"
                                              "QMenu::item:!enabled {color: #a8abb1;}"
                                              "QMenu::item:selected {background-color:#2c9dfc; color:#ffffff;}"));
        }

        // 阴影
        QGraphicsDropShadowEffect  * shadow_effect = new QGraphicsDropShadowEffect(this);
        shadow_effect->setOffset(0,0);
        QColor color("#000000");
        color.setAlphaF(0.2);
        shadow_effect->setColor(color);
        shadow_effect->setBlurRadius(4);
        menu->setGraphicsEffect(shadow_effect);

        if(index.parent().isValid()) {
            // 添加“打开”功能
            action = new PlacesModel::ItemAction(item->index(), tr("Open"), menu);
            connect(action, &QAction::triggered, this, &PlacesView::onOpenTriggered);
            menu->addAction(action);
            if(item->type() != PlacesModelItem::NetServer) { // "连接至服务"右键菜单不需要下面两个功能
                action = new PlacesModel::ItemAction(item->index(), tr("Open in New Tab"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onOpenNewTab);
                menu->addAction(action);
                action = new PlacesModel::ItemAction(item->index(), tr("Open in New Window"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onOpenNewWindow);
                menu->addAction(action);
            }
        }

        switch(item->type()) {
        case PlacesModelItem::Places: {
            auto path = item->path();
            // FIXME: inefficient
            if(path) {
                auto path_str = path.toString();
                if(strcmp(path_str.get(), "trash:///") == 0) {
                    action = new PlacesModel::ItemAction(item->index(), tr("Empty Trash"), menu);
                    auto icn = item->icon();
                    if(icn && icn->qicon().name() == QLatin1String("user-trash")) { // surely an empty trash
                        action->setEnabled(false);
                    }
                    else {
                        connect(action, &QAction::triggered, this, &PlacesView::onEmptyTrash);
                    }
                    // add the "Empty Trash" item on the top
                    QList<QAction*> actions = menu->actions();
                    if(!actions.isEmpty()) {
                        menu->insertAction(actions.at(0), action);
                        menu->insertSeparator(actions.at(0));
                    }
                    else { // impossible
                        menu->addAction(action);
                    }
                }
                // add a "Hide" action to the end
                // menu->addSeparator();
                // action = new PlacesModel::ItemAction(item->index(), tr("Hide"), menu);
                // QString pathStr(path_str.get());
                // action->setCheckable(true);
                // if(proxyModel_->isShowingAll()) {
                //     action->setChecked(proxyModel_->isHidden(pathStr));
                // }
                // connect(action, &QAction::triggered, [this, pathStr](bool checked) {
                //     proxyModel_->setHidden(pathStr, checked);
                //     Q_EMIT hiddenItemSet(pathStr, checked);
                // });
                // menu->addAction(action);
            }
            break;
        }
        case PlacesModelItem::Bookmark: {
            // create context menu for bookmark item
            if(item->index().row() > 0) {
                action = new PlacesModel::ItemAction(item->index(), tr("Move Bookmark Up"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onMoveBookmarkUp);
                menu->addAction(action);
            }
            if(item->index().row() < model_->rowCount()) {
                action = new PlacesModel::ItemAction(item->index(), tr("Move Bookmark Down"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onMoveBookmarkDown);
                menu->addAction(action);
            }
            action = new PlacesModel::ItemAction(item->index(), tr("Rename Bookmark"), menu);
            connect(action, &QAction::triggered, this, &PlacesView::onRenameBookmark);
            menu->addAction(action);
            action = new PlacesModel::ItemAction(item->index(), tr("Remove Bookmark"), menu);
            connect(action, &QAction::triggered, this, &PlacesView::onDeleteBookmark);
            menu->addAction(action);
            break;
        }
        case PlacesModelItem::Volume: {
            PlacesModelVolumeItem* volumeItem = static_cast<PlacesModelVolumeItem*>(item);

            if(volumeItem->isMounted()) {
                action = new PlacesModel::ItemAction(item->index(), tr("Unmount"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onUnmountVolume);
            }
            else {
                action = new PlacesModel::ItemAction(item->index(), tr("Mount"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onMountVolume);
            }
            menu->addAction(action);

            if(volumeItem->canEject() || volumeItem->volumeFromUSB()) {
                action = new PlacesModel::ItemAction(item->index(), tr("Eject"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onEjectVolume);
                menu->addAction(action);
            }
            break;
        }
        case PlacesModelItem::NetConnect: {
            PlacesModelMountItem* volumeItem = static_cast<PlacesModelMountItem*>(item);
            if(volumeItem->isMounted()) {
                action = new PlacesModel::ItemAction(item->index(), tr("Unmount"), menu);
                connect(action, &QAction::triggered, this, &PlacesView::onUnmountMount);
            }
            menu->addAction(action);
            break;
        }
        default:
            break;
        }

        // also add an acton for showing all hidden items
        if(proxyModel_->hasHidden()) {
            if(item->type() == PlacesModelItem::Bookmark) {
                menu->addSeparator();
            }
            action = new PlacesModel::ItemAction(item->index(), tr("Show All Entries"), menu);
            action->setCheckable(true);
            action->setChecked(proxyModel_->isShowingAll());
            connect(action, &QAction::triggered, [this](bool checked) {
                showAll(checked);
            });
            menu->addAction(action);
        }

        if(menu->actions().size()) {
            menu->popup(mapToGlobal(event->pos()));
            connect(menu, &QMenu::aboutToHide, menu, &QMenu::deleteLater);
        }
        else {
            menu->deleteLater();
        }
    }
}

void PlacesView::restoreHiddenItems(const QSet<QString>& items) {
    proxyModel_->restoreHiddenItems(items);
}

void PlacesView::showAll(bool show) {
    proxyModel_->showAll(show);
    if(show) {
        expandAll();
        // for some reason (a Qt bug?), spanning is reset
        spanFirstColumn();
    }
}

void PlacesView::onOpenTriggered() {
    PlacesModel::ItemAction* action = static_cast<PlacesModel::ItemAction*>(sender());
    if(!action->index().isValid()) {
        return;
    }
    PlacesModelItem* item = static_cast<PlacesModelItem*>(model_->itemFromIndex(action->index()));
    if(item) {
        item->updateClickIcon();
        if(item->type() == PlacesModelItem::Volume)
        {
            PlacesModelVolumeItem* itemVolume = static_cast<PlacesModelVolumeItem*>(item);
            if(itemVolume->isMounted() && itemVolume->volumeFromUSB())
            {
                QModelIndex itemIndex = proxyModel_->mapFromSource(item->index());
                QModelIndex btnIndex = itemIndex.sibling(itemIndex.row(), 1);
                QStandardItem* ejectBtn = static_cast<QStandardItem*>(model_->itemFromIndex(proxyModel_->mapToSource(btnIndex)));
                ejectBtn->setIcon(QIcon(":/res/eject-click.png"));
            }
        }
        preItem = item;
        auto path = item->path();
        if(path) {
            Q_EMIT chdirRequested(0, path);
        } else {
            if(item->type() == PlacesModelItem::Type::NetServer){
                Q_EMIT connetServer();
            }
        }
    }

}

bool PlacesView::inRootItem(PlacesModelItem* item)
{
    //return model_->inRootItem(item);
    return true;
}

} // namespace Fm
