/*
 * Copyright (C) 2013 - 2015  Hong Jen Yee (PCMan) <pcman.tw@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "utilities.h"
#include "utilities_p.h"
#include <QApplication>
#include <QClipboard>
#include <QMimeData>
#include <QUrl>
#include <QList>
#include <QStringBuilder>
#include <QMessageBox>
#include "fileoperation.h"
#include <QEventLoop>
#include <QProcess>
#include "mountoperation.h"

#include <pwd.h>
#include <grp.h>
#include <stdlib.h>
#include <glib.h>
#include <QObject>
#include <QStandardPaths>

namespace Fm {

Fm::FilePathList pathListFromUriList(const char* uriList) {
    Fm::FilePathList pathList;
    char** uris = g_strsplit_set(uriList, "\r\n", -1);
    for(char** uri = uris; *uri; ++uri) {
        if(**uri != '\0') {
            pathList.push_back(Fm::FilePath::fromUri(*uri));
        }
    }
    g_strfreev(uris);
    return pathList;
}

QByteArray pathListToUriList(const Fm::FilePathList& paths) {
    QByteArray uriList;
    for(auto& path: paths) {
        uriList += path.uri().get();
        uriList += "\r\n";
    }
    return uriList;
}

Fm::FilePathList pathListFromQUrls(QList<QUrl> urls) {
    Fm::FilePathList pathList;
    for(auto it = urls.cbegin(); it != urls.cend(); ++it) {
        auto path = Fm::FilePath::fromUri(it->toString().toUtf8().constData());
        pathList.push_back(std::move(path));
    }
    return pathList;
}

std::pair<Fm::FilePathList, bool> parseClipboardData(const QMimeData& data) {
    bool isCut = false;
    Fm::FilePathList paths;

    if(data.hasFormat("x-special/gnome-copied-files")) {
        // Gnome, LXDE, and XFCE
        QByteArray gnomeData = data.data("x-special/gnome-copied-files");
        char* pdata = gnomeData.data();
        char* eol = strchr(pdata, '\n');

        if(eol) {
            *eol = '\0';
            isCut = (strcmp(pdata, "cut") == 0 ? true : false);
            paths = pathListFromUriList(eol + 1);
        }
    }

    if(paths.empty() && data.hasUrls()) {
        // The KDE way
        paths = Fm::pathListFromQUrls(data.urls());
        QByteArray cut = data.data(QStringLiteral("application/x-kde-cutselection"));
        if(!cut.isEmpty() && QChar::fromLatin1(cut.at(0)) == QLatin1Char('1')) {
            isCut = true;
        }
    }

    return std::make_pair(paths, isCut);
}

void pasteFilesFromClipboard(const Fm::FilePath& destPath, QWidget* parent) {
    QClipboard* clipboard = QApplication::clipboard();
    const QMimeData* data = clipboard->mimeData();
    Fm::FilePathList paths;
    bool isCut = false;

    std::tie(paths, isCut) = parseClipboardData(*data);

    if(!paths.empty()) {
        if(isCut) {
            FileOperation::moveFiles(paths, destPath, parent);
            clipboard->clear(QClipboard::Clipboard);
        }
        else {
            FileOperation::copyFiles(paths, destPath, parent);
        }
    }
}

void copyFilesToClipboard(const Fm::FilePathList& files) {
    // 禁止复制桌面默认应用文件
    QString filename;
    Fm::FilePathList paths = files;
    QString desktopDir = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    for (int i = 0; i < paths.size(); i++) {
        filename = paths[i].uri().get();
        if( (filename == QObject::tr("file://%1/trash-can.desktop").arg(desktopDir))
            || (filename == QObject::tr("file://%1/user-home.desktop").arg(desktopDir))
            || (filename == QObject::tr("file://%1/computer.desktop").arg(desktopDir))
            || (filename == QObject::tr("file://%1/network.desktop").arg(desktopDir))) {
            paths.erase(paths.begin() + i);
            --i;
        }
    }

    QClipboard* clipboard = QApplication::clipboard();
    QMimeData* data = new QMimeData();
    QByteArray ba;
    auto urilist = pathListToUriList(paths);

    // Add current pid to trace cut/copy operations to current app
    data->setData(QStringLiteral("text/x-libfmqt-pid"), ba.setNum(QCoreApplication::applicationPid()));
    // Gnome, LXDE, and XFCE
    // Note: the standard text/urilist format uses CRLF for line breaks, but gnome format uses LF only
    data->setData("x-special/gnome-copied-files", QByteArray("copy\n") + urilist.replace("\r\n", "\n"));
    // The KDE way
    data->setData("text/uri-list", urilist);
    // data->setData(QStringLiteral("application/x-kde-cutselection"), QByteArrayLiteral("0"));
    clipboard->setMimeData(data);
}

void cutFilesToClipboard(const Fm::FilePathList& files) {
    // 禁止剪切桌面默认应用文件
    QString filename;
    Fm::FilePathList paths = files;
    QString desktopDir = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    for (int i = 0; i < paths.size(); i++) {
        filename = paths[i].uri().get();
        if( (filename == QObject::tr("file://%1/trash-can.desktop").arg(desktopDir))
            || (filename == QObject::tr("file://%1/user-home.desktop").arg(desktopDir))
            || (filename == QObject::tr("file://%1/computer.desktop").arg(desktopDir))
            || (filename == QObject::tr("file://%1/network.desktop").arg(desktopDir))) {
            paths.erase(paths.begin() + i);
            --i;
        }
    }

    QClipboard* clipboard = QApplication::clipboard();
    QMimeData* data = new QMimeData();
    QByteArray ba;
    auto urilist = pathListToUriList(paths);

    // Add current pid to trace cut/copy operations to current app
    data->setData(QStringLiteral("text/x-libfmqt-pid"), ba.setNum(QCoreApplication::applicationPid()));
    // Gnome, LXDE, and XFCE
    // Note: the standard text/urilist format uses CRLF for line breaks, but gnome format uses LF only
    data->setData("x-special/gnome-copied-files", QByteArray("cut\n") + urilist.replace("\r\n", "\n"));
    // The KDE way
    data->setData("text/uri-list", urilist);
    data->setData(QStringLiteral("application/x-kde-cutselection"), QByteArrayLiteral("1"));
    clipboard->setMimeData(data);
}

bool isCurrentPidClipboardData(const QMimeData& data) {
    QByteArray clip_pid = data.data(QStringLiteral("text/x-libfmqt-pid"));
    QByteArray curr_pid;
    curr_pid.setNum(QCoreApplication::applicationPid());

    return !clip_pid.isEmpty() && clip_pid == curr_pid;
}

bool changeFileName(const Fm::FilePath& filePath, const QString& newName, QWidget* parent, bool showMessage) {
    auto dest = filePath.parent().child(newName.toLocal8Bit().constData());
    Fm::GErrorPtr err;
    if(!g_file_move(filePath.gfile().get(), dest.gfile().get(),
                    GFileCopyFlags(G_FILE_COPY_ALL_METADATA |
                                   G_FILE_COPY_NO_FALLBACK_FOR_MOVE |
                                   G_FILE_COPY_NOFOLLOW_SYMLINKS),
                    nullptr, /* make this cancellable later. */
                    nullptr, nullptr, &err)) {
        if (showMessage){
            QMessageBox::critical(parent ? parent->window() : nullptr, QObject::tr("Error"), err.message());
        }
        return false;
    }
    return true;
}

bool renameFile(std::shared_ptr<const Fm::FileInfo> file, QWidget* parent) {
    FilenameDialog dlg(parent ? parent->window() : nullptr);
    dlg.setWindowTitle(QObject::tr("Rename File"));
    dlg.setLabelText(QObject::tr("Please enter a new name:"));
    // FIXME: what's the best way to handle non-UTF8 filename encoding here?
    auto old_name = QString::fromStdString(file->name());
    dlg.setTextValue(old_name);

    if(file->isDir()) { // select filename extension for directories
        dlg.setSelectExtension(true);
    }

    if(dlg.exec() != QDialog::Accepted) {
        return false; // stop multiple renaming
    }

    QString new_name = dlg.textValue();
    if(new_name == old_name) {
        return true; // let multiple renaming continue
    }
    changeFileName(file->path(), new_name, parent);
    return true;
}

bool renameLabel(std::shared_ptr<const Fm::FileInfo> file, QWidget* parent)
{
    if (file->target().empty())
    {
        qInfo() <<"file is empty";
        return false;
    }

    GFile *gfile = nullptr;
    GMount *mount = nullptr;
    GVolume *volume = nullptr;
    gchar *device_name = nullptr;
    char *command = nullptr;
    char *std_output = nullptr;
    int exit_status = 0;
    GError *error_lsblk = nullptr;
    GError *error_command = nullptr;
    GString *cmd = nullptr;
    gboolean res = false;
    bool result = false;
    gfile = g_file_new_for_path(file->target().data());
    if (gfile)
    {
        mount = g_file_find_enclosing_mount(gfile, nullptr,nullptr);
        if (mount)
        {
            volume = g_mount_get_volume(mount);
            if (volume)
            {
                device_name = g_volume_get_identifier(volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                if (device_name)
                {
                    command = g_strdup_printf("lsblk -nlo FSTYPE %s", device_name); // g_strdup_printf will be alway return true, so not need to check command,
                                                                                    // The returned string should be freed with g_free() when no longer needed.
                    res = g_spawn_command_line_sync(command, &std_output, nullptr, &exit_status, &error_lsblk);
                    if (res)
                    {
                        std_output[strlen(std_output)-1] = '\0';
                        FilenameDialog dlg(parent ? parent->window() : nullptr);
                        dlg.setWindowTitle(QObject::tr("Rename Label"));
                        dlg.setLabelText(QObject::tr("Please enter a new name:"));
                        auto old_name = file->displayName();

                        dlg.setTextValue(old_name);

                        if (file->isDir()) // select filename extension for directories
                        {
                            dlg.setSelectExtension(true);
                        }

                        if (dlg.exec() != QDialog::Accepted)
                        {
                            g_free(command);
                            g_free(std_output);
                            g_free(device_name);
                            g_object_unref(volume);
                            g_object_unref(mount);
                            g_object_unref(gfile);
                            return false; // stop multiple renaming, free all resource and return false
                        }

                        QString new_name = dlg.textValue();
                        if (new_name == old_name)
                        {
                            g_free(command);
                            g_free(std_output);
                            g_free(device_name);
                            g_object_unref(volume);
                            g_object_unref(mount);
                            g_object_unref(gfile);
                            return true; // let multiple renaming continue
                        }

                        // use command to rename label
                        if (g_strcmp0(std_output, "ext4") == 0 ||
                            g_strcmp0(std_output, "ext3") == 0 ||
                            g_strcmp0(std_output, "ext2") == 0)
                        {
                            cmd = g_string_new("sudo e2label");
                            g_string_append_printf(cmd," %s %s", device_name, new_name.toStdString().c_str());
                        }
                        else if(g_strcmp0(std_output, "ntfs") == 0)
                        {
                            cmd = g_string_new("sudo ntfslabel --force");
                            g_string_append_printf(cmd," %s %s", device_name, new_name.toStdString().c_str());
                        }
                        else if(g_strcmp0(std_output, "vfat") == 0)
                        {
                            cmd = g_string_new("sudo dosfslabel");
                            g_string_append_printf(cmd," %s %s", device_name, new_name.toStdString().c_str());
                        }
                        else if(g_strcmp0(std_output, "xfs") == 0)
                        {
                            cmd = g_string_new("sudo xfs_admin -fL ");
                            g_string_append_printf(cmd,"%s %s",new_name.toStdString().c_str(),device_name);
                        }
                        else
                        {
                            g_debug("please use gparted to %s",std_output);
                        }

                        if (cmd)
                        {
                            g_spawn_command_line_async(cmd->str, &error_command);
                            qInfo() <<cmd->str;
                            if(!error_command)
                            {
                                result = true;
                            }
                            else
                            {
                                QMessageBox::critical(parent ? parent->window() : nullptr, QObject::tr("Error"), error_command->message);
                                g_error_free(error_command);
                            }
                        }
                        g_string_free(cmd,TRUE);
                    }
                    else
                    {
                        QMessageBox::critical(parent ? parent->window() : nullptr, QObject::tr("Error"), error_lsblk->message);
                        g_error_free(error_lsblk);
                    }
                    //
                    g_free(command);
                    g_free(std_output);
                    g_free(device_name);
                }
                g_object_unref(volume);
            }
            g_object_unref(mount);
        }
        g_object_unref(gfile);
    }
    return result;
}

bool ejectMedia(std::shared_ptr<const Fm::FileInfo> file, QWidget* parent)
{
    if (file->target().empty())
    {
        qInfo() <<"file is empty";
        return false;
    }

    GFile *gfile = nullptr;
    GMount *mount = nullptr;
    GVolume *volume = nullptr;
    GDrive *drive = nullptr;
    GList *ll,*volumes;
    bool result = false;
    gfile = g_file_new_for_path(file->target().data());
    if (gfile)
    {
        mount = g_file_find_enclosing_mount(gfile, nullptr,nullptr);
        if (mount)
        {
            volume = g_mount_get_volume(mount);
            if (volume)
            {
                if (g_volume_can_eject(volume))
                {
                    MountOperation* op = new MountOperation(true, parent);
                    op->eject(volume);
                    op->wait();
                    result = true;
                }
                else
                {
                    gchar *device_name = g_volume_get_identifier(volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
                    char *command = g_strdup_printf("udisksctl power-off -b %s", device_name);
                    drive = g_volume_get_drive(volume);
                    if (drive)
                    {
                        volumes = g_drive_get_volumes(drive);       //get本驱动器上的所有卷设备
                        if (volumes)
                        {
                            for(ll = volumes; ll; ll = ll->next)
                            {
                                GVolume* tempVolume = G_VOLUME(ll->data);
                                char *umountCommand = g_strdup_printf("umount %s", g_volume_get_identifier(tempVolume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE));
                                system(umountCommand);
                                g_object_unref(tempVolume);
                            } //for
                            g_list_free(volumes);
                        }
                        g_object_unref(drive);
                    }
                    system(command);
                }
                g_object_unref(volume);
            }
            g_object_unref(mount);
        }
        g_object_unref(gfile);
    }

    return result;
}

QStringList getIconNames(GThemedIcon *icon)
{
    QStringList iconNames;
    char ** names;
    char ** iter;
    names = nullptr;
    g_object_get(icon,"names",&names,nullptr);
    for (iter = names; *iter; iter++){
        iconNames.append(QString(*iter));
    }
    g_strfreev(names);
    return iconNames;
}

// templateFile is a file path used as a template of the new file.
FilePath createFileOrFolder(CreateFileType type, FilePath parentDir, const TemplateItem* templ, QWidget* parent) {
    QString defaultNewName;
    QString prompt;
    QString dialogTitle = type == CreateNewFolder ? QObject::tr("Create Folder")
                          : QObject::tr("Create File");

    switch(type) {
    case CreateNewTextFile:
        prompt = QObject::tr("Please enter a new file name:");
        defaultNewName = QObject::tr("New text file") + QObject::tr(".txt");
        break;

    case CreateNewFolder:
        prompt = QObject::tr("Please enter a new folder name:");
        defaultNewName = QObject::tr("New folder");
        break;

    case CreateWithTemplate: {
        auto mime = templ->mimeType();
        prompt = QObject::tr("Enter a name for the new %1:").arg(mime->desc());
        defaultNewName = QString::fromStdString(templ->name());
    }
    break;
    }
    int num=1;
    QString new_name = defaultNewName;

_retry:
    // ask the user to input a file name
    // bool ok;
    // QString new_name = QInputDialog::getText(parent ? parent->window() : nullptr,
    //                    dialogTitle,
    //                    prompt,
    //                    QLineEdit::Normal,
    //                    defaultNewName,
    //                    &ok);

    // if(!ok) {
    //     return;
    // }

    auto dest = parentDir.child(new_name.toLocal8Bit().data());
    Fm::GErrorPtr err;
    switch(type) {
    case CreateNewTextFile: {
        Fm::GFileOutputStreamPtr f{g_file_create(dest.gfile().get(), G_FILE_CREATE_NONE, nullptr, &err), false};
        if(f) {
            g_output_stream_close(G_OUTPUT_STREAM(f.get()), nullptr, nullptr);
        }
        break;
    }
    case CreateNewFolder:
        g_file_make_directory(dest.gfile().get(), nullptr, &err);
        break;
    case CreateWithTemplate:
        // copy the template file to its destination
        FileOperation::copyFile(templ->filePath(), dest, parent);
        break;
    }
    if(err) {
        if(err.domain() == G_IO_ERROR && err.code() == G_IO_ERROR_EXISTS) {
            err.reset();
            ++num;
            if(type == CreateWithTemplate) {
                QStringList list = defaultNewName.split(".");
                new_name = list.at(0) + QString("(%1).").arg(num) + list.at(1);
            }
            else if(type == CreateNewTextFile) {
                QStringList list = defaultNewName.split(".");
                new_name = list.at(0) + QString("(%1).").arg(num) + list.at(1);
            }
            else {
                new_name = defaultNewName + QString("%1").arg(num);
            }
            goto _retry;
        }

        QMessageBox::critical(parent ? parent->window() : nullptr, QObject::tr("Error"), err.message());
    }
    return dest;
}

uid_t uidFromName(QString name) {
    uid_t ret;
    if(name.isEmpty()) {
        return INVALID_UID;
    }
    if(name.at(0).digitValue() != -1) {
        ret = uid_t(name.toUInt());
    }
    else {
        struct passwd* pw = getpwnam(name.toLatin1());
        // FIXME: use getpwnam_r instead later to make it reentrant
        ret = pw ? pw->pw_uid : INVALID_UID;
    }

    return ret;
}

QString uidToName(uid_t uid) {
    QString ret;
    struct passwd* pw = getpwuid(uid);

    if(pw) {
        ret = pw->pw_name;
    }
    else {
        ret = QString::number(uid);
    }

    return ret;
}

gid_t gidFromName(QString name) {
    gid_t ret;
    if(name.isEmpty()) {
        return INVALID_GID;
    }
    if(name.at(0).digitValue() != -1) {
        ret = gid_t(name.toUInt());
    }
    else {
        // FIXME: use getgrnam_r instead later to make it reentrant
        struct group* grp = getgrnam(name.toLatin1());
        ret = grp ? grp->gr_gid : INVALID_GID;
    }

    return ret;
}

QString gidToName(gid_t gid) {
    QString ret;
    struct group* grp = getgrgid(gid);

    if(grp) {
        ret = grp->gr_name;
    }
    else {
        ret = QString::number(gid);
    }

    return ret;
}

int execModelessDialog(QDialog* dlg) {
    // FIXME: this does much less than QDialog::exec(). Will this work flawlessly?
    QEventLoop loop;
    QObject::connect(dlg, &QDialog::finished, &loop, &QEventLoop::quit);
    // DialogExec does not seem to be documented in the Qt API doc?
    // However, in the source code of QDialog::exec(), it's used so let's use it too.
    dlg->show();
    (void)loop.exec(QEventLoop::DialogExec);
    return dlg->result();
}

// check if GVFS can support this uri scheme (lower case)
// NOTE: this does not work reliably due to some problems in gio/gvfs and causes bug lxqt/lxqt#512
// https://github.com/lxqt/lxqt/issues/512
// Use uriExists() whenever possible.
bool isUriSchemeSupported(const char* uriScheme) {
    const gchar* const* schemes = g_vfs_get_supported_uri_schemes(g_vfs_get_default());
    if(Q_UNLIKELY(schemes == nullptr)) {
        return false;
    }
    for(const gchar * const* scheme = schemes; *scheme; ++scheme)
        if(strcmp(uriScheme, *scheme) == 0) {
            return true;
        }
    return false;
}

// check if the URI exists.
// NOTE: this is a blocking call possibly involving I/O.
// So it's better to use it in limited cases, like checking trash:// or computer://.
// Avoid calling this on a slow filesystem.
// Checking "network:///" is very slow, for example.
bool uriExists(const char* uri) {
    GFile* gf = g_file_new_for_uri(uri);
    bool ret = (bool)g_file_query_exists(gf, nullptr);
    g_object_unref(gf);
    return ret;
}

QString formatFileSize(uint64_t size, bool useSI) {
    Fm::CStrPtr str{g_format_size_full(size, useSI ? G_FORMAT_SIZE_DEFAULT : G_FORMAT_SIZE_IEC_UNITS)};
    return QString(str.get());
}

char * get_home_dir (void)
{
  struct passwd *pw;
  static char *home_dir = NULL;
  const char *home_env;

  if (home_dir != NULL)
    return home_dir;

  home_env = getenv ("HOME");

  if (home_env && *home_env)
    {
      home_dir = strdup (home_env);
    }
  else
    {
      setpwent ();
      pw = getpwuid (getuid ());
      endpwent ();

      if (pw && pw->pw_dir)
        home_dir = strdup (pw->pw_dir);
    }

  return home_dir;
}

char * get_user_config_file (const char *filename)
{
  char *config_home;
  char * file = static_cast<char*>(malloc(128)) ;
  int free_config_home;
  qInfo()<<getenv ("XDG_CONFIG_HOME");
  config_home = getenv ("XDG_CONFIG_HOME");
  free_config_home = 0;
  if (config_home == NULL || config_home[0] == 0)
    {
      config_home = strcat (get_home_dir (), "/.config");
      free_config_home = 1;
    }
  strcpy(file,config_home);
  strcat(file,"/");
  strcat(file,filename);

  if (free_config_home)
    free (config_home);

  return file;
}
int is_space (char c)
{
  return c == ' ' || c == '\t';
}
int has_prefix (const char *str, const char *prefix)
{
  return strncmp (str, prefix, strlen (prefix)) == 0;
}

bool renameuserfile(std::shared_ptr<const Fm::FileInfo>  file,const QString& newName)
{
    FILE * config_file;
    char buffer[512];
    char *user_config_file;
    char *p,*key, *key_end,*value, *value_end;
    int len;
    char * homepath;
    char con_filename[512]="";
    bool result = false;
    std::string filename = file->path().toString().get();

    user_config_file = get_user_config_file ("user-dirs.dirs");

    config_file = fopen (user_config_file, "r");


    if (config_file == NULL)
      return false;

    while (fgets (buffer, sizeof (buffer), config_file))
    {
        /* Remove newline at end */
        len = strlen (buffer);
        if (len > 0 && buffer[len-1] == '\n')
            buffer[len-1] = 0;

        p = buffer;
        /* Skip whitespace */
        while (is_space (*p))
            p++;

        if (*p == '#')
            continue;

        if (!has_prefix(p, "XDG_"))
            continue;
        p += 4;
        key = p;

        while (*p && !is_space (*p) && * p != '=')
            p++;

        key_end = p - 4;
        if (key_end <= key ||
            !has_prefix (key_end, "_DIR"))
            continue;


        if (*p == '=')
            p++;

        while (is_space (*p))
            p++;

        if (*p++ != '"')
            continue;


        if (has_prefix (p, "$HOME"))
        {
            p += 5;
            if (*p == '/')
                p++;
            else if (*p != '"' && *p != 0)
                continue;
            homepath = getenv("HOME");
        }
        else if (*p != '/')
            continue;
        else homepath = "/";

        value = p;

        while (*p)
        {
            if (*p == '"')
              break;
            if (*p == '\\' && *(p+1) != 0)
              p++;

            p++;
        }
        value_end = p;
        *value_end = 0;

        char * name = strdup(value);
        memset(con_filename,0,512);
        strcat(con_filename,homepath);
        strcat(con_filename,"/");
        strcat(con_filename,name);

        if(strcmp(con_filename,filename.c_str()) ==0)
        {
            qInfo()<<filename.c_str()<<" same";
            QString desktopPath = user_config_file;
            QStringList program;
            program << "-c" << QString("sed -i 's/%1/%2/g' %3").arg(name).arg(newName).arg(desktopPath);
            QProcess *systemProcess = new QProcess();
            systemProcess->start("/bin/bash", program);
            systemProcess->waitForFinished();
            systemProcess->close();
            result = true;
            break;
        }
    }
    free (user_config_file);
    fclose(config_file);
    return result;
}

QStringList getmenutype(const char * type)
{
    FILE * config_file;
    char buffer[512];
    char *p,*key, *key_end;
    int len;
    QStringList list;
    config_file = fopen ("/usr/share/applications/mimeinfo.cache", "r");
    if (config_file == NULL)
      return list;

    while (fgets (buffer, sizeof (buffer), config_file))
    {
        len = strlen (buffer);
        if (len > 0 && buffer[len-1] == '\n')
            buffer[len-1] = 0;

        p = buffer;
        if (!has_prefix(p, type))
            continue;
        key = p;

        while (*p && !is_space (*p) && * p != '=')
            p++;
        key_end =p;
        *key_end = 0;
        char * keyname = strdup(key);
        list.push_back(keyname);
    }
    fclose(config_file);
    return  list;
}


} // namespace Fm
